"""
Show what's going on inside an input filter.
"""
import numpy as np

import becca_viz.viz_tools as vt


def render(filter, bbox, y_pool, pool_viz_map, radius=0):
    """
    Make a picture of the discretization that happens in an input filter.

    To simplify the notation on transforms here, the following substitutions
    will be used:
        A:  feature pool in visualized order
        B:  feature pool in natural order (the order passed in)
        C:  features in natural order
        D:  features in visualized order

    Parameters
    ----------
    bbox: list of floats
        Of the form [x_bottom, x_top, y_left, y_right]
    filter: InputFilter
    y_pool: array of floats
    pool_viz_map: 2D array of ints
    radius: float

    Returns
    -------
    y_features: array of floats
    feature_viz_map: 2D array of ints
    """
    xmin, xmax, ymin, ymax = bbox
    # frame_width = xmax - xmin
    frame_height = ymax - ymin

    n_pool = y_pool.size
    y_pool_spacing = (frame_height - 2 * radius) / (n_pool + 1)
    y_A = np.linspace(
        ymin + radius + y_pool_spacing,
        ymax - radius - y_pool_spacing,
        num=n_pool,
        endpoint=True,
    )

    map_AB = pool_viz_map.T
    activities_B = filter.candidate_activities
    activities_A = np.matmul(map_AB, activities_B)

    n_inputs = (np.where(np.sum(filter.mapping, axis=0))[0]).size
    map_BC = filter.mapping[:n_pool, :n_inputs]
    map_AC = np.matmul(map_AB, map_BC)
    order_CD = np.argsort(np.argsort(np.matmul(y_A, map_AC)))
    map_CD = np.zeros((n_inputs, n_inputs), dtype=np.int)
    map_CD[np.arange(n_inputs, dtype=np.int), order_CD] = 1
    map_AD = np.matmul(map_AC, map_CD)

    y_spacing = (frame_height - 2 * radius) / (n_inputs + 1)
    y_D = np.linspace(
        ymin + radius + y_spacing,
        ymax - radius - y_spacing,
        num=n_inputs,
        endpoint=True,
    )
    activities_D = np.matmul(activities_A, map_AD)

    i_DA = np.matmul(np.arange(n_pool, dtype=np.int), map_AD)
    # Show the filter's selection
    for i_D, activity in enumerate(activities_D):
        i_A = i_DA[i_D]
        y_end = y_D[i_D]
        y_start = y_A[i_A]
        vt.plot_curve_activity_horiz(y_start, y_end, xmin, xmax, activity)
        vt.plot_point_activity(xmax, y_end, activity, y_spacing)

    return y_D, map_CD



"""
Show what's going on inside the postprocessor.
"""
import numpy as np

import becca_viz.viz_tools as vt


def render(postprocessor, bbox, radius=0, phase=0):
    """
    Make a picture of the discretization that happens in the Preprocessor.

    Parameters
    ----------
    bbox: list of floats
        Of the form [x_bottom, x_top, y_left, y_right]
    phase: int
    postprocessor: Postprocessor
    radius: float
        The corner radius of the box.

    Returns
    -------
    x_commands: array of floats
        The absolute x positions of the command nodes.
    i_to_viz: 2D array of ints
        Mapping from commands to visualization order.
    """

    xmin, xmax, ymin, ymax = bbox
    frame_width = xmax - xmin
    # frame_height = ymax - ymin
    n_commands = postprocessor.n_commands
    n_actions = postprocessor.n_actions
    # Collect positions from all the commands.
    n_commands_per_action = int(n_commands / n_actions)
    x_command_spacing = (
        (frame_width - 2 * radius) / (n_commands + 1))
    x_commands = (xmin + radius + x_command_spacing *
                  np.cumsum(np.ones(n_commands)))
    x_action_spacing = (frame_width - 2 * radius) / (n_actions + 1)
    x_actions = (xmin + radius + x_action_spacing *
                 np.cumsum(np.ones(n_commands)))

    i_to_viz = np.eye(n_commands, dtype=np.int)

    # Show the upward, sensed commands from the previous time step.
    for i_action, _ in enumerate(postprocessor.actions):
        for j in range(n_commands_per_action):
            i_command = i_action * n_commands_per_action + j
            command_activity = postprocessor.previous_commands[
                i_command]

            vt.plot_point_activity(
                x_commands[i_command],
                ymax,
                command_activity,
                x_command_spacing,
            )
            vt.plot_curve_activity(
                x_actions[i_action],
                x_commands[i_command],
                ymin,
                ymax,
                command_activity,
            )

    offset = (x_actions[1] - x_actions[0]) / 10
    # Show the downward commands, to be executed during the next time step.
    for i_action, action in enumerate(postprocessor.actions):
        for j in range(n_commands_per_action):
            i_command = i_action * n_commands_per_action + j
            command_activity = postprocessor.command_activities[
                i_command]
            if command_activity > .02:
                vt.plot_point_activity(
                    x_commands[i_command] + offset,
                    ymax,
                    command_activity,
                    x_command_spacing,
                    activity_color=vt.oxide,
                )
                vt.plot_curve_activity(
                    x_actions[i_action] + offset,
                    x_commands[i_command] + offset,
                    ymin,
                    ymax,
                    command_activity,
                    activity_color=vt.oxide,
                )
                vt.plot_point_activity(
                    x_actions[i_action] + offset,
                    ymin,
                    action,
                    x_action_spacing,
                    activity_color=vt.oxide,
                )
    return x_commands, i_to_viz



import os

import numpy as np
import matplotlib.patches as patches
import matplotlib.pyplot as plt

import becca_viz.viz_tools as vt


def visualize(affect, brain, verbose=True):
    """
    Parameters
    ----------
    affect: Affect
    brain: Brain
    """
    # Plot the lifetime record of the reward.
    fig = plt.figure(11111)
    fig.clf()
    ax = plt.gca()

    color = (np.array(vt.copper) +
             np.random.normal(size=3, scale=.1))
    color = np.maximum(np.minimum(color, 1), 0)
    color = tuple(color)
    linewidth = np.random.normal(loc=2.5)
    linewidth = 2
    linewidth = np.maximum(1, linewidth)
    plt.plot(
        np.array(affect.reward_steps) / 1000.,
        affect.reward_history,
        color=color,
        linewidth=linewidth,
    )

    x_min = 0
    x_max = affect.reward_steps[-1] / 1000
    y_min = np.minimum(np.min(affect.reward_history), 0)
    y_max = np.maximum(np.max(affect.reward_history), 1)
    ax.add_patch(patches.Rectangle(
        (x_min, y_min),
        x_max - x_min,
        y_max - y_min,
        facecolor=vt.copper_highlight,
        edgecolor='none',
        zorder=-16,
    ))
    plt.axis([x_min, x_max, y_min, y_max])

    plt.xlabel('Thousands of time steps')
    plt.ylabel('Average reward')
    plt.title('Reward history for {0}'.format(brain.name))

    # Save a copy of the plot.
    filename = 'reward_history_{0}.png'.format(brain.name)
    pathname = os.path.join(brain.log_dir, filename)
    plt.savefig(pathname, format='png')
    if verbose:
        print('Saved', pathname)
    return



import numpy as np

import becca_viz.viz_tools as vt


def render(featurizer, bbox, viz_maps, radius=0):
    """

    Parameters
    ----------
    featurizer : Featurizer
    bbox: list of floats
        In the format [xmin, xmax, ymin, ymax]
    viz_maps: list of 2D arrays of ints
        Maps between cable candidate pools and their visualization order.
    radius: float

    Returns
    -------
    y_pool_feature: array of floats
        The absolute positions of each member of the feature pool.
    feature_viz_map: 2D array of floats
        Map between the feature pool and their visualization order.
    """
    xmin, xmax, ymin, ymax = bbox
    # frame_width = xmax - xmin
    frame_height = ymax - ymin

    if viz_maps[-1] is None:
        viz_maps = viz_maps[:-1]
    block_rows = []
    for i_map in np.arange(len(viz_maps), dtype=np.int):
        block_row = []
        mrows, mcols = viz_maps[i_map].shape
        for j_map in np.arange(len(viz_maps), dtype=np.int):
            nrows, ncols = viz_maps[j_map].shape
            if i_map == j_map:
                block_row.append(np.fliplr(viz_maps[i_map]))
            else:
                block_row.append(np.zeros((mrows, ncols), dtype=np.int))
        block_rows.append(block_row)

    map_AB = np.block(block_rows).T
    map_BC = featurizer.mapping
    order_CD = np.argsort(np.argsort(np.matmul(
        np.arange(map_AB.shape[0]),
        np.matmul(map_AB, map_BC))))
    n_D = map_BC.shape[1]
    map_CD = np.zeros((n_D, n_D), dtype=np.int)
    map_CD[np.arange(n_D, dtype=np.int), order_CD] = 1

    activities_B = []
    for level_activities in featurizer.activities:
        activities_B += list(level_activities)
    activities_B = np.array(activities_B)
    activities_D = np.matmul(activities_B, np.matmul(map_BC, map_CD))

    y_spacing = (frame_height - 2 * radius) / (n_D + 1)
    y_D = np.linspace(
        ymin + radius + y_spacing,
        ymax - radius - y_spacing,
        num=n_D,
        endpoint=True,
    )

    for i_D, activity in enumerate(activities_D):
        # vt.plot_curve_activity_horiz(
        #     y_start, y_end, xmin, xmax, activity, start=.3)
        vt.plot_point_activity(xmin, y_D[i_D], activity, y_spacing)

    return y_D, map_CD



"""
Show what's going on inside the preprocessor.
"""
import numpy as np

import becca_viz.viz_tools as vt


def render(preprocessor, bbox, radius=0):
    """
    Make a picture of the discretization that happens in the Preprocessor.

    Parameters
    ----------
    bbox: list of floats
        Of the form [x_bottom, x_top, y_left, y_right]
    preprocessor: Preprocessor

    Returns
    -------
    input_x_viz: array of floats
        The absolute x positions of the input nodes.
    input_viz_map: 2D array of ints
        A map from the order the inputs occur to the visualization order.
    """
    xmin, xmax, ymin, ymax = bbox
    frame_width = xmax - xmin
    frame_height = ymax - ymin

    n_inputs = preprocessor.n_inputs
    n_disc = len(preprocessor.discretizers)
    # The number of inputs plus the number of discretizers.
    n_input_d = n_inputs + n_disc
    positions = np.zeros(n_input_d)

    x_spacing = (frame_width - 2 * radius) / float(len(positions) + 1)
    node_x = xmin + radius + x_spacing * np.cumsum(np.ones(len(positions)))

    # Get the x positions of all the nodes.
    # This needs to be in the same order as activities.
    i_disc = 0
    for i_disc, discretizer in enumerate(preprocessor.discretizers):
        positions[i_disc] = discretizer.position
        num_tree_list = discretizer.numeric_cats.get_list()
        str_tree_list = discretizer.string_cats.get_list()
        for node in num_tree_list:
            positions[node.i_input + n_disc] = node.position
        for node in str_tree_list:
            positions[node.i_input + n_disc] = node.position

    position_order = np.argsort(positions)
    i_input_d_by_position = np.arange(n_input_d)[position_order]
    x_by_i_input_d = node_x[np.argsort(i_input_d_by_position)]
    discretizer_x = x_by_i_input_d[:n_disc]
    input_x = x_by_i_input_d[n_disc:]
    input_viz_map = np.zeros((n_inputs, n_inputs), dtype=np.int)
    input_viz_map[np.argsort(input_x), np.arange(n_inputs)] = 1
    input_x_viz = np.matmul(input_x, input_viz_map)

    for i_discretizer, discretizer in enumerate(preprocessor.discretizers):
        # Build trees.
        n_depth = 1. + np.maximum(
            discretizer.numeric_cats.depth,
            discretizer.string_cats.depth)
        branch_length = frame_height / n_depth
        sensor_x = discretizer_x[i_discretizer]

        def plot_tree(tree):
            root_x = input_x[tree.root.i_input]
            activity = preprocessor.input_activities[tree.root.i_input]
            plot_branch(sensor_x, root_x, ymin, activity, branch_length)
            nodes = tree.get_list()
            for node in nodes:
                if node.lo_child is not None:
                    y_start = ymin + (node.depth + 1.) * branch_length
                    x_start = input_x[node.i_input]
                    x_end = input_x[node.lo_child.i_input]
                    lo_activity = preprocessor.input_activities[
                        node.lo_child.i_input]
                    plot_branch(
                        x_start,
                        x_end,
                        y_start,
                        lo_activity,
                        branch_length,
                        is_leaf=node.lo_child.leaf,
                        max_y=ymax,
                    )
                    vt.plot_point_activity(
                        x_end,
                        y_start + branch_length,
                        lo_activity,
                        x_spacing,
                    )
                    x_end = input_x[node.hi_child.i_input]
                    hi_activity = preprocessor.input_activities[
                        node.hi_child.i_input]
                    plot_branch(
                        x_start,
                        x_end,
                        y_start,
                        hi_activity,
                        branch_length,
                        is_leaf=node.hi_child.leaf,
                        max_y=ymax,
                    )
                    vt.plot_point_activity(
                        x_end,
                        y_start + branch_length,
                        hi_activity,
                        x_spacing,
                    )

        # Build string branch.
        plot_tree(discretizer.string_cats)
        plot_tree(discretizer.numeric_cats)

    return input_x_viz, input_viz_map


def plot_branch(
    x_start, x_end,
    y_start,
    activity,
    branch_length,
    is_leaf=False,
    max_y=0.,
):
    """
    Draw the branches of the discretized sensor trees.

    @param y_start, x_start, x_end: floats
        The start and end coordinates of the branch.
    @param activity: float
        The activity level of the branch, between 0 and 1.
    @param branch_length: float
        The y distance between generations in the tree.
    @param is_leaf: boolean
        Is this branch for a leaf node?
    @param max_y: float
        In the case of a leaf node, what is the x extent of the window.
    """
    y_end = y_start + branch_length
    vt.plot_curve_activity(x_start, x_end,  y_start, y_end, activity)
    if is_leaf:
        vt.plot_line_activity([x_end, x_end], [y_end, max_y], activity)



import matplotlib.pyplot as plt
import numpy as np
import matplotlib.colors as colors
import matplotlib.patches as patches

# Colors for plotting
dark_grey = (0.05, 0.05, 0.05)
light_grey = (0.9, 0.9, 0.9)
red = (0.9, 0.3, 0.3)
# Becca pallette
copper_highlight = (253/255, 249/255, 240/255)
light_copper = (242/255, 166/255, 108/255)
copper = (175/255, 102/255, 53/255)
dark_copper = (132/255, 73/255, 36/255)
copper_shadow = (25/255, 22/255, 20/255)
oxide = (20/255, 120/255, 150/255)

cmap_1 = colors.LinearSegmentedColormap.from_list(
    "", ["black", light_copper])
cmap_2 = colors.LinearSegmentedColormap.from_list(
    "", [oxide, "black", light_copper])

xlabel_shift = 3 / 4
ylabel_shift = 1 / 2
title_shift = 1 / 100
text_shift = 1 / 20


def draw_frame(
    bbox=None,
    radius=.01,
    edgecolor=copper,
    facecolor=None,
    linewidth=1,
):
    """
    Draw a frame with rounded corners.

    Parameters
    ----------
    bbox: list of floats
        Of the form [x_left, x_right, y_bottom, y_top]
    radius: float
        The radius of the rounded corners
    edgecolor, facecolor: pyplot color specifier
    linewidth: float
        Thickness of border line.
    """
    ax = plt.gca()
    xmin, xmax, ymin, ymax = bbox
    width = xmax - xmin
    height = ymax - ymin
    z_frame = -8
    z_fill = -10

    # Fill in the background.
    if facecolor is not None:
        # Fill the tall part of the frame.
        ax.add_patch(patches.Rectangle(
            (xmin + radius, ymin),
            width - 2 * radius,
            height,
            facecolor=facecolor,
            edgecolor='none',
            zorder=z_fill,
        ))
        # Fill the wide part of the frame.
        ax.add_patch(patches.Rectangle(
            (xmin, ymin + radius),
            width,
            height - 2 * radius,
            facecolor=facecolor,
            edgecolor='none',
            zorder=z_fill,
        ))
        # Fill the corners.
        ax.add_patch(patches.Circle(
            (xmin + radius, ymin + radius),
            radius=radius,
            facecolor=facecolor,
            edgecolor='none',
            zorder=z_fill,
        ))
        ax.add_patch(patches.Circle(
            (xmin + radius, ymax - radius),
            radius=radius,
            facecolor=facecolor,
            edgecolor='none',
            zorder=z_fill,
        ))
        ax.add_patch(patches.Circle(
            (xmax - radius, ymin + radius),
            radius=radius,
            facecolor=facecolor,
            edgecolor='none',
            zorder=z_fill,
        ))
        ax.add_patch(patches.Circle(
            (xmax - radius, ymax - radius),
            radius=radius,
            facecolor=facecolor,
            edgecolor='none',
            zorder=z_fill,
        ))

    # Draw the frame.
    # Right wall
    plt.plot(
        [xmin + radius, xmax - radius],
        [ymax, ymax],
        color=edgecolor,
        linewidth=linewidth,
        zorder=z_frame,
    )
    # Left wall
    plt.plot(
        [xmin + radius, xmax - radius],
        [ymin, ymin],
        color=edgecolor,
        linewidth=linewidth,
        zorder=z_frame,
    )
    # Bottom wall
    plt.plot(
        [xmin, xmin],
        [ymin + radius, ymax - radius],
        color=edgecolor,
        linewidth=linewidth,
        zorder=z_frame,
    )
    # Top wall
    plt.plot(
        [xmax, xmax],
        [ymin + radius, ymax - radius],
        color=edgecolor,
        linewidth=linewidth,
        zorder=z_frame,
    )
    # Upper right corner
    plot_arc(
        color=edgecolor,
        linewidth=linewidth,
        radius=radius,
        theta_0=0,
        theta_1=np.pi / 2,
        x_center=xmax - radius,
        y_center=ymax - radius,
        zorder=z_frame,
    )
    # Upper left corner
    plot_arc(
        color=edgecolor,
        linewidth=linewidth,
        radius=radius,
        theta_0=np.pi / 2,
        theta_1=np.pi,
        x_center=xmin + radius,
        y_center=ymax - radius,
        zorder=z_frame,
    )
    # Lower right corner
    plot_arc(
        color=edgecolor,
        linewidth=linewidth,
        radius=radius,
        theta_0=3 * np.pi / 2,
        theta_1=2 * np.pi,
        x_center=xmax - radius,
        y_center=ymin + radius,
        zorder=z_frame,
    )
    # Lower left corner
    plot_arc(
        color=edgecolor,
        linewidth=linewidth,
        radius=radius,
        theta_0=np.pi,
        theta_1=3 * np.pi / 2,
        x_center=xmin + radius,
        y_center=ymin + radius,
        zorder=z_frame,
    )


def plot_arc(
    color='green',
    linewidth=1,
    radius=1,
    theta_0=0,
    theta_1=np.pi/2,
    x_center=0,
    y_center=0,
    zorder=0,
):
    """
    Put an arc in the current axes.

    Parameters
    ----------
    radius: float
    theta_0, theta_1: floats
        The start and end angles of the arc, in radians.
        theta_0 must be smaller than theta_1
    x_center, y_center: floats
        The coordinates of the center of the arc.
    z_order: int
        The over/under-lap priority for the arc.
    """
    ds = .001
    # dtheta = ds / radius
    theta = np.arange(theta_0, theta_1, ds)
    x = x_center + np.cos(theta) * radius
    y = y_center + np.sin(theta) * radius

    plt.plot(x, y, color=color, linewidth=linewidth, zorder=zorder)


def plot_point_activity(
    x, y,
    activity,
    spacing,
    activity_color=light_copper,
):
    """
    Draw a point that represents the activity of a signal it carries.
    """
    r = np.minimum(np.maximum(.01, spacing / 24.), .03)
    circle = plt.Circle(
        (x, y),
        r,
        color=copper_shadow,
        zorder=6.,
    )
    plt.gca().add_artist(circle)
    r_activity = activity * r * 2.
    circle = plt.Circle(
        (x, y),
        r_activity,
        color=activity_color,
        zorder=7. + activity,
    )
    plt.gca().add_artist(circle)


def plot_curve_activity(
    x_start, x_end, y_start, y_end,
    activity,
    shadow_color=copper_shadow,
    activity_color=light_copper,
):
    """
    Draw a smooth curve connecting two points.
    """
    t = np.arange(0., np.pi, .01)
    curve = np.cos(t)
    offset = (x_start + x_end) / 2.
    dilation = (y_end - y_start) / np.pi
    scale = (
        np.sign(y_start - y_end) *
        (x_end - x_start) / 2.
    )
    plt.plot(
        offset + curve * scale,
        y_start + t * dilation,
        color=shadow_color,
        linewidth=.8,
        zorder=2.,
    )
    plt.plot(
        offset + curve * scale,
        y_start + t * dilation,
        color=activity_color,
        linewidth=activity,
        zorder=activity + 2.,
    )


def plot_curve_activity_horiz(
    y_start, y_end, x_start, x_end,
    activity,
    shadow_color=copper_shadow,
    activity_color=light_copper,
):
    """
    Draw a smooth curve connecting two points.
    """
    t = np.arange(0., np.pi, .01)
    curve = np.cos(t)
    offset = (y_start + y_end) / 2.
    dilation = (x_end - x_start) / np.pi
    scale = (
        np.sign(x_start - x_end) *
        (y_end - y_start) / 2.
    )
    plt.plot(
        x_start + t * dilation,
        offset + curve * scale,
        color=shadow_color,
        linewidth=.8,
        zorder=2.,
    )
    plt.plot(
        x_start + t * dilation,
        offset + curve * scale,
        color=activity_color,
        linewidth=activity,
        zorder=activity + 2.,
    )


def plot_line_activity(x, y, activity):
    """
    Draw a line that represents the activity of a signal it carries.
    """

    plt.plot(
        x, y,
        color=copper_shadow,
        linewidth=.8,
        zorder=2.,
    )
    plt.plot(
        x, y,
        color=light_copper,
        linewidth=activity,
        zorder=activity + 2.,
    )


def scatter(x, y, c):
    cb = c.copy()
    cb[:, 3] *= cb[:, 3]
    cc = cb.copy()
    cc[:, 3] *= cc[:, 3]
    plt.scatter(x, y, c="black", marker='.', s=15, edgecolor='none', zorder=1)
    plt.scatter(x, y, c=cc, marker='.', s=7, edgecolor='none', zorder=2)
    plt.scatter(x, y, c=cb, marker='.', s=4, edgecolor='none', zorder=3)
    plt.scatter(x, y, c=c, marker='.', s=2, edgecolor='none', zorder=4)


def scatter_1D(
    arr,
    x0=0,
    y0=0,
    width=1,
    xlabel='',
    title='',
):
    """
    Make a 1D scatter plot.

    Put it in a row.
    """
    npts = arr.size
    x = np.zeros(npts)
    y = np.zeros(npts)
    c = np.zeros((npts, 4))
    c[:, :3] = np.array(light_copper)
    x_unit = width / (npts - 1)

    for i_x in range(npts):
        x[i_x] = i_x * x_unit
        val = np.maximum(np.minimum(arr[i_x], 1), -1)
        if val > 0:
            c[i_x, 3] = val
        else:
            c[i_x, :3] = np.array(oxide)
            c[i_x, 3] = -val

    x += x0
    y += y0
    scatter(x, y, c)

    text_sep = width * text_shift
    plt.text(
        x0 + width * xlabel_shift,
        y0 - text_sep,
        xlabel,
        fontsize=4,
        color=copper,
        verticalalignment="top",
        horizontalalignment="left",
        family="sans-serif",
    )
    plt.text(
        x0 + width * title_shift,
        y0 + text_sep,
        title,
        fontsize=6,
        color=copper,
        verticalalignment="bottom",
        horizontalalignment="left",
        family="sans-serif",
    )
    return


def scatter_2D(
    arr,
    x0=0,
    y0=0,
    width=1,
    height=1,
    xlabel='',
    ylabel='',
    title='',
):
    """
    Make a 2D scatter plot.

    row number corresponds to y value
    column number corresponds to x value
    """
    nrows, ncols = arr.shape
    npts = nrows * ncols
    x = np.zeros(npts)
    y = np.zeros(npts)
    c = np.zeros((npts, 4))
    c[:, :3] = np.array(light_copper)
    y_unit = height / (nrows - 1)
    x_unit = width / (ncols - 1)

    for i_x in range(ncols):
        for i_y in np.arange(nrows - 1, -1, -1, dtype=np.int):
                j = i_x * nrows + i_y
                x[j] = i_x * x_unit
                y[j] = i_y * y_unit
                val = np.maximum(np.minimum(arr[i_y, i_x], 1), -1)
                if val > 0:
                    c[j, 3] = val
                else:
                    c[j, :3] = np.array(oxide)
                    c[j, 3] = -val

    x += x0
    y += y0
    scatter(x, y, c)

    text_sep = np.minimum(height, width) * text_shift
    plt.text(
        x0 + width * xlabel_shift,
        y0 - text_sep,
        xlabel,
        fontsize=4,
        color=copper,
        verticalalignment="top",
        horizontalalignment="left",
        family="sans-serif",
    )
    plt.text(
        x0 - text_sep,
        y0 + height * ylabel_shift,
        ylabel,
        fontsize=4,
        color=copper,
        rotation=-90,
        verticalalignment="center",
        horizontalalignment="right",
        family="sans-serif",
    )
    plt.text(
        x0 + width * title_shift,
        y0 + height + text_sep,
        title,
        fontsize=6,
        color=copper,
        verticalalignment="bottom",
        horizontalalignment="left",
        family="sans-serif",
    )
    return


def scatter_3D(
    arr,
    x0=0,
    y0=0,
    width=1,
    height=1,
    xlabel='goals',
    ylabel='features',
    zlabel='outcomes',
    title='Active sequences',
):
    """
    """
    nrows, ncols, ndeps = arr.shape
    npts = nrows * ncols * ndeps
    x = np.zeros(npts)
    y = np.zeros(npts)
    c = np.zeros((npts, 4))
    c[:, :3] = np.array(light_copper)

    center = np.array([x0 + width / 2, y0 + height / 2])
    # Direction vectors with foreshortening and scaling
    fs = [.9, .7, 1]
    scale = .6
    length = np.sqrt(width * height)
    # theta
    x_dir = fs[0] * scale * np.array([
        np.cos(-10 * np.pi / 180), np.sin(-10 * np.pi / 180)])
    y_dir = fs[1] * scale * np.array([
        np.cos(30 * np.pi / 180), np.sin(30 * np.pi / 180)])
    z_dir = fs[2] * scale * np.array([0, 1])
    # Unit vectors
    x_unit = x_dir * length / ncols
    y_unit = y_dir * length / nrows
    z_unit = z_dir * length / ndeps
    # Origin
    x_0 = -x_unit * ncols / 2
    y_0 = -y_unit * nrows / 2
    z_0 = -z_unit * ndeps / 2
    xy_0 = center + x_0 + y_0 + z_0

    x = np.zeros(npts)
    y = np.zeros(npts)
    c = np.zeros((npts, 4))
    c[:, :3] = np.array(light_copper)

    for i_x in range(ncols):
        for i_y in np.arange(nrows - 1, -1, -1, dtype=np.int):
            for i_z in range(ndeps):
                j = i_x * nrows * ndeps + i_y * ndeps + i_z
                x_part = i_x * x_unit
                y_part = i_y * y_unit
                z_part = i_z * z_unit
                x[j], y[j] = x_part + y_part + z_part
                c[j, 3] = np.minimum(arr[i_y, i_x, i_z], 1)
    x += xy_0[0]
    y += xy_0[1]
    scatter(x, y, c)

    plt.text(
        x0 + .22 * width,
        y0 + .08 * height,
        xlabel,
        fontsize=4,
        color=copper,
        rotation=-10,
        verticalalignment="top",
        horizontalalignment="left",
        family="sans-serif",
    )
    plt.text(
        x0 + .05 * width,
        y0 + .4 * height,
        ylabel,
        fontsize=4,
        color=copper,
        rotation=-90,
        verticalalignment="center",
        horizontalalignment="right",
        family="sans-serif",
    )
    plt.text(
        x0 + .7 * width,
        y0 + .15 * height,
        zlabel,
        fontsize=4,
        color=copper,
        rotation=30,
        verticalalignment="top",
        horizontalalignment="left",
        family="sans-serif",
    )
    plt.text(
        x0 + width * title_shift,
        y0 + height * .86,
        title,
        fontsize=6,
        color=copper,
        verticalalignment="bottom",
        horizontalalignment="left",
        family="sans-serif",
    )
    return


def nd_map(arr, map):
    """
    Apply a map to a multidimensional array.

    Parameters
    ----------
    arr: nD array of floats
    map: 2D array of ints

    Returns
    -------
    nD array of floats
        Re-mapped
    """
    nD = len(arr.shape)
    dim_list = list(np.arange(nD, dtype=np.int))
    shift_list = list(dim_list)
    shift_list.append(shift_list.pop(0))

    map_size = map.shape[0]
    if nD == 1:
        arr = arr[:map_size]
    elif nD == 2:
        arr = arr[:map_size, :map_size]
    elif nD == 3:
        arr = arr[:map_size, :map_size, :map_size]
    elif nD == 4:
        arr = arr[:map_size, :map_size, :map_size, :map_size]

    for _ in range(nD):
        arr = np.matmul(arr, map)
        arr = np.moveaxis(arr, dim_list, shift_list)

    return arr



"""
Show the world and what's going on inside the brain.
"""
import os

import numpy as np
import matplotlib.patches as patches
import matplotlib.pyplot as plt

import becca_viz.viz_tools as vt
import becca_viz.preprocessor_viz as preprocessor_viz
import becca_viz.postprocessor_viz as postprocessor_viz
import becca_viz.input_filter_viz as input_filter_viz
import becca_viz.ziptie_viz as ziptie_viz
import becca_viz.featurizer_viz as featurizer_viz
import becca_viz.feature_filter_viz as feature_filter_viz
import becca_viz.model_viz as model_viz


lwf = .3  # Linewidth of the frames

wd_ = 18  # Total width of the image
ht_ = 9  # Total heigh of the image
brd = .75  # Border thickness around frames

halfwidth_weight = 5
centerwidth_weight = 1
total_weight = 2 * halfwidth_weight + centerwidth_weight

no_borders = wd_ - brd * 3  # Width available after borders are removed
wdh = no_borders * halfwidth_weight / total_weight  # Width of wide columns
wdc = no_borders * centerwidth_weight / total_weight  # Center column width

htf = ht_ - brd * 2  # Full height, minus borders
n_zipties = 2  # Assumed number of zipties
n_rows = 1 + 2 * n_zipties  # Total number of frames stacked vertically
no_borders_ht = ht_ - brd * (n_zipties + 2)  # Total non-border height
htr = no_borders_ht / n_rows  # height of each row
wdq = (wdh - brd) / 2  # Quarter-width
wds = (wdh - brd) / 3  # Sixth-width

rad = htr / 6  # The radius of the rounded corners on frames

preprocessor_bbox = [brd, brd + 2 * wds, brd, brd + htr]
postprocessor_bbox = [2 * brd + 2 * wds, brd + wdh, brd, brd + htr]
filt_0_bbox = [brd, brd + wdh, 2 * brd + htr, 2 * brd + 2 * htr]
ziptie_0_bbox = [brd, brd + wdh, 2 * brd + 2 * htr, 2 * brd + 3 * htr]
filt_1_bbox = [brd, brd + wdh, 3 * brd + 3 * htr, 3 * brd + 4 * htr]
ziptie_1_bbox = [brd, brd + wdh, 3 * brd + 4 * htr, 3 * brd + 5 * htr]
feature_filt_bbox = [2 * brd + wdh, 2 * brd + wdh + wdc, brd, brd + htf]
model_bbox = [
    2 * brd + wdh + wdc, 2 * brd + 2 * wdh + wdc, brd, brd + htf]


def visualize(brain):
    """
    Render the sensor information making its way up through the brain.

    Parameters
    ----------
    brain: Brain
    """
    # TODO: Incorporate this into the main visualization.
    brain.affect.visualize(brain)
    create_background()
    x_inputs, preprocessor_viz_map = preprocessor_viz.render(
        brain.preprocessor, preprocessor_bbox, radius=rad)
    x_commands, postprocessor_viz_map = postprocessor_viz.render(
        brain.postprocessor, postprocessor_bbox, radius=rad)
    n_pre = preprocessor_viz_map.shape[0]
    n_post = postprocessor_viz_map.shape[0]

    pool_0_viz_map = np.block([
        [np.zeros((n_post, n_pre), dtype=np.int), postprocessor_viz_map],
        [preprocessor_viz_map, np.zeros((n_pre, n_post), dtype=np.int)]])
    # TODO: handle multiple zipties
    x_pool_0 = np.concatenate((x_inputs, x_commands))
    x_cables_0, cables_0_viz_map = input_filter_viz.render(
        brain.featurizer.filter,
        filt_0_bbox,
        x_pool_0,
        pool_0_viz_map,
        preprocessor_bbox[3],  # max y value of the Preprocessor
        radius=rad)
    x_pool_1, pool_1_viz_map = ziptie_viz.render(
        brain.featurizer.ziptie,
        ziptie_0_bbox,
        x_cables_0,
        cables_0_viz_map,
        filt_0_bbox[3],  # max y value of the InputFilter
        radius=rad)
    pool_viz_map = [pool_0_viz_map, pool_1_viz_map]
    y_pool_feature, feature_pool_viz_map = featurizer_viz.render(
        brain.featurizer,
        feature_filt_bbox,
        pool_viz_map,
        radius=rad)
    y_feature, feature_viz_map = feature_filter_viz.render(
        brain.model.filter,
        feature_filt_bbox,
        y_pool_feature,
        feature_pool_viz_map,
        radius=rad)
    model_viz.render(
        brain.model,
        brain.actor,
        model_bbox,
        feature_viz_map,
        radius=rad)

    finalize(brain, dpi=300)


def labels(brain):
    """
    Generate a set of labels for the visualization.
    """
    create_background()
    title_frame("Preprocessor", preprocessor_bbox, radius=rad)
    title_frame("Postprocessor", postprocessor_bbox, radius=rad)
    title_frame("Cable filter 0", filt_0_bbox, radius=rad)
    title_frame("Ziptie 0", ziptie_0_bbox, radius=rad)
    title_frame("Cable filter 1", filt_1_bbox, radius=rad)
    title_frame("Ziptie 1", ziptie_1_bbox, radius=rad)
    title_frame("Feature\nfilter", feature_filt_bbox, radius=rad)
    title_frame("Model", model_bbox, radius=rad)
    model_viz.labels(model_bbox, radius=rad)

    finalize(brain, dpi=300, tag='labels')


def title_frame(title, bbox, radius=0):
    """
    Label the rendered components.
    """
    xmin, xmax, ymin, ymax = bbox
    plt.text(
        xmin + radius,
        ymax - radius,
        title,
        fontsize=16,
        color=vt.copper,
        verticalalignment="top",
        family="sans-serif",
    )


def create_background(edgecolor=vt.oxide, facecolor=vt.dark_grey):
    """
    Set up the backdrop for the visualization.
    """
    fig = plt.figure(num=84782, figsize=(wd_, ht_), frameon=False)
    fig.clf()
    ax = plt.gca()
    ax.add_patch(patches.Rectangle(
        (0, 0),
        wd_,
        ht_,
        facecolor=facecolor,
        edgecolor='none',
        zorder=-16.,
    ))
    # Preprocessor frame
    vt.draw_frame(
        bbox=preprocessor_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )
    # Postprocessor frame
    vt.draw_frame(
        bbox=postprocessor_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )
    # Cable filter 0 frame
    vt.draw_frame(
        bbox=filt_0_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )
    # Ziptie 0 frame
    vt.draw_frame(
        bbox=ziptie_0_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )
    # Cable filter 1 frame
    vt.draw_frame(
        bbox=filt_1_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )
    # Ziptie 1 frame
    vt.draw_frame(
        bbox=ziptie_1_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )
    # feature filter frame
    vt.draw_frame(
        bbox=feature_filt_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )
    # model frame
    vt.draw_frame(
        bbox=model_bbox,
        radius=rad,
        facecolor=facecolor,
        edgecolor=edgecolor,
    )

    return


def finalize(brain, dpi=300, tag='activity', verbose=True):
    """
    Complete any final formatting and save a copy of the figure.

    Parameters
    ----------
    brain: Brain
    dpi: int
        The dots per inch for the saved figure.
    phase: int
        During each time step the brain does a lot and it's tough to get
        it all into one image. To handle this, it is broken up into phases.
        Phase 1 is sensing (an upward pass through the architecture) and
        Phase 2 is acting ( a downward pass through the architecture).

    """
    plt.tight_layout()
    plt.axis('equal')
    plt.axis('off')

    filename = 'becca_{name}_{dpi:04d}_{age:08d}_{tag}.png'.format(
        name=brain.name, age=brain.timestep, dpi=dpi, tag=tag)
    pathname = os.path.join(brain.log_dir, filename)
    plt.savefig(pathname, format='png', dpi=dpi)
    if verbose:
        print('Saved', pathname)



import matplotlib.pyplot as plt
import numpy as np

import becca_viz.viz_tools as vt


n_image_rows = 4
n_image_cols = 3


def render(model, actor, bbox, viz_map, radius=0):
    """
    Turn it into a picture.
    """
    xmin, xmax, ymin, ymax = bbox
    frame_width = xmax - xmin
    frame_height = ymax - ymin

    y_gap = 2 * radius
    im_height = (frame_height - (n_image_rows + 1) * y_gap) / n_image_rows
    x_gap = (frame_width - n_image_cols * im_height) / (n_image_cols + 1)

    rewards = (model.feature_activities[:, np.newaxis] *
               model.prefix_rewards)[2:, 2:]
    viz_rewards = vt.nd_map(rewards, viz_map)
    autoscaled_viz_rewards = viz_rewards / (
        np.max(np.abs(viz_rewards)) + 1e-2)
    vt.scatter_2D(
        autoscaled_viz_rewards,
        x0=xmin + 2 * x_gap + im_height,
        y0=(ymin + (n_image_rows - 1) * y_gap
            + (n_image_rows - 1) * im_height),
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='features',
        title='Active rewards',
    )

    curiosities = (model.feature_activities[:, np.newaxis]
                   * model.prefix_curiosities)[2:, 2:]
    viz_curiosities = vt.nd_map(curiosities, viz_map)
    autoscaled_viz_curiosities = viz_curiosities / (
        np.max(viz_curiosities) + 1e-2)
    vt.scatter_2D(
        autoscaled_viz_curiosities,
        x0=xmin + x_gap,
        y0=(ymin + (n_image_rows - 1) * y_gap
            + (n_image_rows - 1) * im_height),
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='features',
        title='Active curiosities',
    )

    # TODO: weight all by feature activities
    sequences = (model.feature_activities[:, np.newaxis, np.newaxis] *
                 model.sequence_likelihoods)
    sequences_viz = vt.nd_map(sequences[2:, 2:, 2:], viz_map)
    # Goal direction is x.
    # Post feature direction is y.
    viz_futures = np.max(sequences_viz, axis=0).transpose()
    # sequences = np.moveaxis(sequences, [0, 1, 2], [1, 2, 0])

    prefix_activities = vt.nd_map(model.prefix_activities[2:, 2:], viz_map)
    vt.scatter_2D(
        prefix_activities,
        x0=xmin + x_gap,
        y0=(ymin + (n_image_rows - 2) * y_gap
            + (n_image_rows - 2) * im_height),
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='features',
        title='Prefix activities',
    )

    vt.scatter_2D(
        viz_futures,
        x0=xmin + 2 * x_gap + im_height,
        y0=(ymin + (n_image_rows - 2) * y_gap
            + (n_image_rows - 2) * im_height),
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='outcomes',
        title='Futures',
    )

    vt.scatter_3D(
        sequences_viz,
        x0=xmin + x_gap,
        y0=ymin + y_gap,
        width=2 * im_height + x_gap,
        height=2 * im_height + y_gap,
        xlabel='goals',
        ylabel='features',
        zlabel='outcomes',
        title='Active\nsequences',
    )

    conditional_rewards_viz = vt.nd_map(
        model.conditional_rewards[2:], viz_map)
    vt.scatter_1D(
        conditional_rewards_viz,
        x0=xmin + 3 * x_gap + 2 * im_height,
        y0=ymin + 4 * y_gap + 3.5 * im_height,
        width=im_height,
        xlabel='goals',
        title='Conditional rewards',
    )

    conditional_curiosities_viz = vt.nd_map(
        model.conditional_curiosities[2:], viz_map)
    vt.scatter_1D(
        conditional_curiosities_viz,
        x0=xmin + 3 * x_gap + 2 * im_height,
        y0=ymin + 3 * y_gap + 2.5 * im_height,
        width=im_height,
        xlabel='goals',
        title='Conditional curiosities',
    )

    conditional_goal_rewards = np.max(
        model.conditional_predictions *
        model.goal_activities[np.newaxis, :], axis=1)
    conditional_goal_rewards_viz = vt.nd_map(
        conditional_goal_rewards[2:], viz_map)
    vt.scatter_1D(
        conditional_goal_rewards_viz,
        x0=xmin + 3 * x_gap + 2 * im_height,
        y0=ymin + 1 * y_gap + 0.5 * im_height,
        width=im_height,
        xlabel='goals',
        title='Conditional goal rewards',
    )

    goal_activities_viz = vt.nd_map(
        actor.goal_collection[2:], viz_map)
    # model.goal_activities[2:], viz_map)
    vt.scatter_1D(
        goal_activities_viz,
        x0=xmin + 3 * x_gap + 2 * im_height,
        y0=ymin + 2 * y_gap + 1.5 * im_height,
        width=im_height,
        xlabel='goals',
        title='Goal collection',
    )
    return


def labels(bbox, radius=0):
    """
    Add model-specific labels to the labeled visualization.
    """
    xmin, xmax, ymin, ymax = bbox
    frame_width = xmax - xmin
    frame_height = ymax - ymin

    y_gap = radius
    im_height = (frame_height - (n_image_rows + 1) * y_gap) / n_image_rows
    x_gap = (frame_width - n_image_cols * im_height) / (n_image_cols + 1)

    label_text(
        text='Active\nreward\nprefixes',
        x=xmin + 2 * x_gap + im_height,
        y=ymin + (n_image_rows - 0) * y_gap + (n_image_rows - .5) * im_height,
    )

    label_text(
        text='Active\ncuriosity\nprefixes',
        x=xmin + x_gap,
        y=ymin + (n_image_rows - 0) * y_gap + (n_image_rows - .5) * im_height,
    )

    label_text(
        text='Prefix\nactivities',
        x=xmin + x_gap,
        y=ymin + (n_image_rows - 1) * y_gap + (n_image_rows - 1.5) * im_height,
    )

    label_text(
        text='Active\nfutures',
        x=xmin + 2 * x_gap + im_height,
        y=ymin + (n_image_rows - 1) * y_gap + (n_image_rows - 1.5) * im_height,
    )

    label_text(
        text='Active\nsequences',
        x=xmin + x_gap,
        y=ymin + 2 * y_gap + im_height,
    )

    label_text(
        text='Conditional\nrewards',
        x=xmin + 3 * x_gap + 2 * im_height,
        y=ymin + 4 * y_gap + 3.5 * im_height,
    )

    label_text(
        text='Conditional\ncuriosities',
        x=xmin + 3 * x_gap + 2 * im_height,
        y=ymin + 3 * y_gap + 2.5 * im_height,
    )

    label_text(
        text='Goal\nactivities',
        x=xmin + 3 * x_gap + 2 * im_height,
        y=ymin + 2 * y_gap + 1.5 * im_height,
    )

    label_text(
        text='Goal\nrewards',
        x=xmin + 3 * x_gap + 2 * im_height,
        y=ymin + 1 * y_gap + 0.5 * im_height,
    )


def label_text(text='', x=0, y=0):
    """
    Craft the label text.
    """
    plt.text(
        x,
        y,
        text,
        fontsize=12,
        color=vt.copper,
        verticalalignment="center",
        family="sans-serif",
    )


def render_state(model, bbox, viz_map, radius=0):
    """

    """
    xmin, xmax, ymin, ymax = bbox
    frame_width = xmax - xmin
    frame_height = ymax - ymin

    i_viz = np.matmul(viz_map, np.arange(viz_map.shape[0], dtype=np.int))
    # Handle the model's two internal features.
    i_viz = np.concatenate((np.array([0, 1], dtype=np.int), i_viz + 2))

    y_gap = radius
    n_image_rows = 4
    im_height = (frame_height - (n_image_rows + 1) * y_gap) / n_image_rows
    x_gap = (frame_width - 3 * im_height) / 3

    vt.scatter_2D(
        model.prefix_rewards[i_viz, :][:, i_viz],
        x0=xmin + 2 * x_gap + im_height,
        y0=ymin + (n_image_rows - 1) * y_gap + (n_image_rows - 1) * im_height,
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='features',
        title='Prefix rewards',
    )

    vt.scatter_2D(
        model.prefix_curiosities[i_viz, :][:, i_viz],
        x0=xmin + x_gap,
        y0=ymin + (n_image_rows - 1) * y_gap + (n_image_rows - 1) * im_height,
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='features',
        title='Prefix curiosities',
    )

    # TODO: weight all by feature activities
    sequences = model.sequence_occurrences / (model.prefix_occurrences + 1)
    # Goal direction is x.
    # Post feature direction is y.
    futures = np.max(sequences, axis=0)

    vt.scatter_2D(
        model.prefix_uncertainties[i_viz, :][:, i_viz],
        x0=xmin + x_gap,
        y0=ymin + (n_image_rows - 2) * y_gap + (n_image_rows - 2) * im_height,
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='features',
        title='Prefix uncertainties',
    )

    vt.scatter_2D(
        futures[i_viz, :][:, i_viz],
        x0=xmin + 2 * x_gap + im_height,
        y0=ymin + (n_image_rows - 2) * y_gap + (n_image_rows - 2) * im_height,
        width=im_height,
        height=im_height,
        xlabel='goals',
        ylabel='outcomes',
        title='Futures',
    )

    sequences_viz = np.moveaxis(
        sequences[i_viz, :, :][:, i_viz, :][:, :, i_viz],
        [0, 1, 2], [2, 1, 0])
    vt.scatter_3D(
        sequences_viz,
        x0=xmin + x_gap,
        y0=ymin + y_gap,
        width=2 * im_height + x_gap,
        height=2 * im_height + y_gap,
    )



# import os

# import matplotlib.pyplot as plt
import numpy as np

import becca_viz.viz_tools as vt


def render(ziptie, bbox, x_inputs, cable_viz_map, y_prev, radius=0):
    """
    Make a picture of the cable-to-bundle combinations that happen
    in a ziptie.

    To keep notation simpler, maps will be referred to as map_xy,
    where x and y can be any of the following:
        A:  cable visualization order
        B:  cable natural order (in which the ziptie receives it)
        C:  bundle natural order (in which ziptie.map creates it)
        D:  bundle visualization order

    Parameters
    ----------
    bbox: list of floats
        Of the form [x_left, x_right, y_bottom, y_top]
    ziptie: Ziptie
    cable_viz_map: 2D array of ints
    y_prev: float
    radius: float

    Returns
    -------
    x_bundles: array of floats
        The absolute x positions of the cables.
    bundle_viz_map: 2D array of ints
        The indices for ordering bundles to align with
        the visualization.
    """
    xmin, xmax, ymin, ymax = bbox
    frame_width = xmax - xmin
    # frame_height = ymax - ymin
    n_A = x_inputs.size
    n_D = ziptie.n_bundles

    x_A_spacing = (frame_width - 2 * radius) / (n_A + 1)
    x_A = np.linspace(
        xmin + radius + x_A_spacing,
        xmax - radius - x_A_spacing,
        num=n_A,
        endpoint=True,
    )

    map_BA = cable_viz_map
    map_AB = map_BA.T
    activities_B = ziptie.cable_activities[:n_A]
    # activities_A = np.matmul(map_AB, activities_B)
    # activities_A = np.matmul(activities_B, map_BA)

    if n_D > 0:
        x_D_spacing = (frame_width - 2 * radius) / (n_D + 1)
        x_D = np.linspace(
            xmin + radius + x_D_spacing,
            xmax - radius - x_D_spacing,
            num=n_D,
            endpoint=True,
        )

        map_BC = ziptie.mapping[:n_A, :n_D]
        x_B = np.matmul(x_A, map_AB)
        bundle_score = (np.sum(x_B[:, np.newaxis] * map_BC, axis=0) /
                        np.sum(map_BC, axis=0))
        map_CD = np.zeros((n_D, n_D), dtype=np.int)
        map_CD[np.arange(n_D, dtype=np.int),
               np.argsort(np.argsort(bundle_score))] = 1
        i_CD = np.matmul(map_CD, np.arange(n_D, dtype=np.int))
        i_BA = np.matmul(np.arange(n_A, dtype=np.int), map_AB)

        for i_B, activity in enumerate(activities_B):
            i_A = i_BA[i_B]
            for i_C in np.where(map_BC[i_B, :])[0]:
                i_D = i_CD[i_C]
                x_end = x_D[i_D]
                x_start = x_A[i_A]
                vt.plot_curve_activity(x_start, x_end, ymin, ymax, activity)
        activities_C = ziptie.bundle_activities
        activities_D = np.matmul(activities_C, map_CD)
        for i_D, activity in enumerate(activities_D):
            vt.plot_point_activity(x_D[i_D], ymax, activity, x_D_spacing)
    else:
        x_D = None
        map_CD = None

    x_bundles = x_D
    bundle_viz_map = map_CD
    return x_bundles, bundle_viz_map



from setuptools import setup

setup(
    name='becca_viz',
    version='0.10.0',
    description='Visualization tools for Becca',
    url='http://github.com/brohrer/becca_viz',
    download_url='https://github.com/brohrer/becca_viz/archive/master.zip',
    author='Brandon Rohrer',
    author_email='brohrer@gmail.com',
    license='MIT',
    packages=['becca_viz'],
    include_package_data=True,
    install_requires=[
        'matplotlib',
    ],
    zip_safe=False)



"""
Show what's going on inside an input filter.
"""
import numpy as np

import becca_viz.viz_tools as vt


def render(filter, bbox, x_pool_prev, pool_viz_map, y_prev,  radius=0):
    """
    Make a picture of the discretization that happens in an input filter.

    To simplify the notation on transforms here, the following substitutions
    will be used:
        A:  candidate pool in visualized order
        B:  candidate pool in natural order (the order passed in)
        C:  inputs in natirual order
        D:  inputs in visualized order

    Parameters
    ----------
    bbox: list of floats
        Of the form [x_bottom, x_top, y_left, y_right]
    filter: InputFilter
    radius: float

    Returns
    -------
    x_cables: array of floats
        The absolute x positions of the cables.
    i_to_viz_cables: 2D array of ints
        Mapping from cable pool to their visualization order
    """
    xmin, xmax, ymin, ymax = bbox
    frame_width = xmax - xmin
    # frame_height = ymax - ymin

    n_pool = x_pool_prev.size
    x_pool_spacing = (frame_width - 2 * radius) / (x_pool_prev.size + 1)
    x_A = np.linspace(
        xmin + radius + x_pool_spacing,
        xmax - radius - x_pool_spacing,
        num=n_pool,
        endpoint=True,
    )

    map_BA = pool_viz_map
    map_AB = map_BA.T
    activities_B = filter.candidate_activities
    activities_A = np.matmul(activities_B, map_BA)

    # Connect the previous block(s) to this one.
    for i_A, activity in enumerate(activities_A):
        vt.plot_curve_activity(
            x_pool_prev[i_A], x_A[i_A],
            y_prev, ymin, activity)
        vt.plot_point_activity(
            x_A[i_A], ymin, activity, x_pool_spacing)

    n_inputs = (np.where(np.sum(filter.mapping, axis=0))[0]).size
    map_BC = filter.mapping[:n_pool, :n_inputs]
    map_AC = np.matmul(map_AB, map_BC)
    order_CD = np.argsort(np.argsort(np.matmul(x_A, map_AC)))
    map_CD = np.zeros((n_inputs, n_inputs), dtype=np.int)
    map_CD[np.arange(n_inputs, dtype=np.int), order_CD] = 1
    map_AD = np.matmul(map_AC, map_CD)

    x_spacing = (frame_width - 2 * radius) / (n_inputs + 1)
    x_D = np.linspace(
        xmin + radius + x_spacing,
        xmax - radius - x_spacing,
        num=n_inputs,
        endpoint=True,
    )
    activities_D = np.matmul(activities_A, map_AD)

    i_DA = np.matmul(np.arange(n_pool, dtype=np.int), map_AD)
    # Show the filter's selection
    for i_D, activity in enumerate(activities_D):
        i_A = i_DA[i_D]
        x_end = x_D[i_D]
        x_start = x_A[i_A]
        vt.plot_curve_activity(x_start, x_end, ymin, ymax, activity)
        vt.plot_point_activity(x_end, ymax, activity, x_spacing)

    return x_D, map_CD
